/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: emulator_c_iface.c,v 1.7 2004/07/28 21:15:42 cheako Exp $
 *
 * Description: Functions for the C emulating core.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "emulator_c_iface.h"
#include "mapper.h"
#include "globals.h"
#include "types.h"
#include "emulator.h"
#include "emulator_c.h"
#include "mapper.h"
#include "controller.h"
#include "renderer.h"
#include "io.h"

signed int INRET;
unsigned int VRAMPTR;
signed int CTNI;
static int last_clock = 0; /* For vblank bit */
extern unsigned int irqflag;
extern int vbl;
int ESI = 0;

static void mmc1_shim_m6502(int addr, int val)
{
    int cesi = ESI;
    cesi = cesi - CTNI + CLOCK;
    CTNI = ESI;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    mmc1(addr, val);
    ESI = CTNI;
}

static void mmc3_shim_m6502(int addr, int val)
{
    int cesi = ESI;
    cesi = cesi - CTNI + CLOCK;
    CTNI = ESI;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    mmc3(addr, val);
    ESI = CTNI;
}

static void aorom_shim_m6502(int addr, int val)
{
    int cesi = ESI;
    cesi = cesi - CTNI + CLOCK;
    CTNI = ESI;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    aorom(addr, val);
    ESI = CTNI;
}

static void input_shim_m6502(int addr)
{
    int cesi = ESI;
    cesi = cesi - CTNI + CLOCK;
    CTNI = ESI;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    /*      printf("input_shim %04x",addr); */
    input(addr);
    ESI = CTNI;			/* unnecessary, input doesn't change CTNI at this time */
}

static void output_shim_m6502(int addr, int val)
{
    int cesi = ESI;
    cesi = cesi - CTNI + CLOCK;
    CTNI = ESI;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    /* exit(1); */
    output(addr, val);
    ESI = CTNI;			/* output() may change CTNI */
}

static int nmi_shim_m6502(int addr)
{
    int cesi;
    int donmi_ret;
    cesi = 7 - CTNI + CLOCK;
    if (cesi - CPF >= 0) {
	cesi -= CPF;
    }
    CLOCK = cesi;
    /* Multiple exit points are confusing */
    if (irqflag) {
	/* Normal IRQ */
	CTNI = CLOCK - VBL;
	irqflag = 0;
	/* No IRQ if background is off */
	if (!((*(unsigned char *) (REG1 + 1)) & 0x08))
	    return INT_NONE;
	return INT_IRQ;
    } else {
	/* NMI */
	donmi_ret = DoNMIC();
	if (donmi_ret == INT_NONE) {
	    return INT_NONE;
	} else if (donmi_ret == INT_QUIT) {
	    return INT_QUIT;
	}
    }
    return INT_NMI;
}

void InitMemoryC(void)
{
    if ((emulator->Mapper = malloc ((MAXMAPPER + 1) * sizeof (void *))) == NULL) {
        perror ("malloc");
	exit (1);
    }
    memset (emulator->Mapper, 0x00, MAXMAPPER + 1);
    if ((emulator->MapperInit = malloc ((MAXMAPPER + 1) * sizeof (void *))) == NULL) {
        perror ("malloc");
	exit (1);
    }
    memset (emulator->MapperInit, 0x00, MAXMAPPER + 1);
    if ((emulator->ROM = malloc (0x300000)) == NULL) {
        perror ("malloc");
	exit (1);
    }
    memset (emulator->ROM, 0x00, 0x300000);
    if ((emulator->RAM = malloc (0x8000)) == NULL) {
        perror ("malloc");
	exit (1);
    }
    memset (emulator->RAM, 0x00, 0x8000);
}

void FreeMemoryC(void)
{
    free (emulator->Mapper);
    free (emulator->ROM);
    free (emulator->RAM);
}

int DoNMIC(void)
{
    /*printf("donmi: at %d\n",CLOCK); */
    CLOCK = VBL + 7;		/* 7 cycle interrupt latency */
    CTNI = -CPF + 7;
    if (last_clock >= CLOCK || last_clock <= 27393)
	vbl = 1;
    last_clock = CLOCK;
    /*quit =*/ renderer->UpdateDisplay();
    /*if (quit) {
	return INT_QUIT;
    }*/
    controller->HandleJoystick(0);
    controller->HandleJoystick(1);
    /*printf("donmi: stack at %x\n",STACKPTR); */
    /* reset scroll registers */
    /*hvscroll = 0; */
    /* Is an NMI to be generated? */
    if (((*((unsigned char *) (REG1))) & 0x80) != 0)
	return INT_NMI;
    else
	return INT_NONE;
}

void mnone(void) {}

void InitMapperSubsystemC(void)
{
    int x;
    /* clear them all to zero */
    for (x = 0; x < MAXMAPPER; x++) {
	emulator->MapperInit[x] = 0;
	emulator->Mapper[x] = 0;
	MapperName[x][0] = '\0';
    }
    emulator->MapperInit[0] = &init_none;
    emulator->Mapper[0] = &mnone;
    strcpy(MapperName[0], "No mapper");
    emulator->MapperInit[1] = &init_mmc1;
    emulator->Mapper[1] = &mmc1_shim_m6502;
    strcpy(MapperName[1], "MMC1");
    emulator->MapperInit[2] = &init_unrom;
    emulator->Mapper[2] = &unrom;
    strcpy(MapperName[2], "UNROM");
    emulator->MapperInit[3] = &init_cnrom;
    emulator->Mapper[3] = &cnrom;
    strcpy(MapperName[3], "CNROM");
    emulator->MapperInit[4] = &init_mmc3;
    emulator->Mapper[4] = &mmc3_shim_m6502;
    strcpy(MapperName[4], "MMC3");
    emulator->MapperInit[5] = &init_mmc5;
    emulator->Mapper[5] = &mmc5;
    strcpy(MapperName[5], "MMC5");
    emulator->MapperInit[7] = &init_aorom;
    emulator->Mapper[7] = &aorom_shim_m6502;
    strcpy(MapperName[7], "AOROM");
    emulator->MapperInit[9] = &init_mmc2;
    emulator->Mapper[9] = &mmc2;
    strcpy(MapperName[9], "MMC2");
    emulator->MapperInit[11] = &init_clrdrms;
    emulator->Mapper[11] = &clrdrms;
    strcpy(MapperName[11], "Color Dreams");
    emulator->MapperInit[13] = &init_cprom;
    emulator->Mapper[13] = &cprom;
    strcpy(MapperName[13], "CPROM");
    emulator->MapperInit[15] = &init_100in1;
    emulator->Mapper[15] = &m100in1;
    strcpy(MapperName[15], "100-in-1");
    emulator->MapperInit[19] = &init_namcot106;
    emulator->Mapper[19] = &namcot106;
    strcpy(MapperName[19], "Namcot 106");
    emulator->MapperInit[22] = &init_vrc2_a;
    emulator->Mapper[22] = &vrc2_a;
    strcpy(MapperName[22], "Konami VRC2 type A");
    emulator->MapperInit[23] = &init_vrc2_b;
    emulator->Mapper[23] = &vrc2_b;
    strcpy(MapperName[23], "Konami VRC2 type B");
    emulator->MapperInit[32] = &init_g101;
    emulator->Mapper[32] = &g101;
    strcpy(MapperName[32], "Irem G-101");
    emulator->MapperInit[33] = &init_taito_tc0190;
    emulator->Mapper[33] = &taito_tc0190;
    strcpy(MapperName[33], "Taito TC0190");
    emulator->MapperInit[64] = &init_tengen_rambo1;
    emulator->Mapper[64] = &tengen_rambo1;
    strcpy(MapperName[64], "Tengen RAMBO-1");
    emulator->MapperInit[66] = &init_gnrom;
    emulator->Mapper[66] = &gnrom;
    strcpy(MapperName[66], "GNROM");
    emulator->MapperInit[68] = &init_sunsoft4;
    emulator->Mapper[68] = &sunsoft4;
    strcpy(MapperName[68], "Sunsoft Mapper #4");
    emulator->MapperInit[69] = &init_fme7;
    emulator->Mapper[69] = &fme7;
    strcpy(MapperName[69], "Sunsoft FME-7");
    emulator->MapperInit[71] = &init_camerica;
    emulator->Mapper[71] = &camerica;
    strcpy(MapperName[71], "Camerica");
    emulator->MapperInit[78] = &init_irem_74hc161_32;
    emulator->Mapper[78] = &irem_74hc161_32;
    strcpy(MapperName[78], "Irem 74HC161/32");
    emulator->MapperInit[99] = &init_vs;
    emulator->Mapper[99] = &vs;
    strcpy(MapperName[99], "VS UniSystem");
    emulator->MapperInit[225] = &init_supervision;
    emulator->Mapper[225] = &supervision;
    strcpy(MapperName[225], "SuperVision");
    emulator->MapperInit[231] = &init_nina7;
    emulator->Mapper[231] = &nina7;
    strcpy(MapperName[231], "NINA-07");
}

void mem_write(uint32_t Addr, uint8_t Value) 
{
	void ((*fxn)(int,int));
//	printf("Wr6502 at %04x (%02x) (CLOCK %d)\n", Addr, Value, CLOCK);
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
	if (Addr & 0x8000) { 
		/* Redirect addresses $8000-$FFFF through mapper. */
		/* asm linkage addr is stored in emulator->Mapper[MAPPERNUM];
		 * from there asm calls C mmc1, mmc3, etc. */
		fxn = emulator->Mapper[MAPPERNUMBER];
		/* Can't remember how to cast to void fxn(int,int) ;) */
		fxn(Addr,Value);
	} else if (Addr >= 0x2000 && Addr < 0x6000) {
	    /* Redirect addresses $2000-$5FFF to Output. */
		/* table.x86 redirects are inconsistent. */
		output_shim_m6502(Addr,Value);
	} else {
		/* RAM */
		/* (ZPMEM and NVRAM are contained in RAM, so I'm omitting the
		 * extra tests for convenience.) */
		emulator->RAM[Addr] = Value;
	}
}

uint8_t mem_read(uint32_t Addr) 
{
	int ret;
// if (Addr > 0xffff) { printf("Addr>0xffff\n"); quit(); }
//   printf("Rd6502 at %04x (CLOCK %d) ->", Addr, CLOCK);
//       printf("MAPTABLE[%d]=%p  ->", Addr>>12,MAPTABLE[Addr>>12]+Addr);
	if (Addr >= 0x2000 && Addr < 0x6000) {
		input_shim_m6502(Addr);
//		printf(" -> %02x\n", INRET & 0xFF);
		return INRET & 0xFF;
	} else {
		ret=*(MAPTABLE[Addr>>12] + Addr); 
//		printf(" %02x\n", ret);
		return ret;
	}
}

void StartC(void) 
{
	int nmi;

	ram = emulator->RAM;  // or MAPTABLE[0]?
	stack_page = ram + STACK_OFFSET; // or STACK?
	nes6502_init();
	nes6502_reset();

	for (;;) {
		ESI += nes6502_execute(100);
//		printf(";\n");

		if (ESI >= 0) {
			nmi = nmi_shim_m6502(reg_PC);
			if (nmi == INT_NMI) {
//				printf("NMI\n");
				nes6502_nmi();   // And nmi executes at top of next _execute.
			} else if (nmi == INT_IRQ) {
				nes6502_irq();
			} else if (nmi == INT_QUIT) {
				break;
			}
			ESI = CTNI;
		}
	}
}
