/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: renderer.c,v 1.18 2004/07/30 04:25:03 cheako Exp $
 *
 * Description: This file provides the "auto" and "none"
 * pseudo-renderers.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <sys/stat.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "consts.h"
#include "globals.h"
#include "mapper.h"
#include "renderer.h"
#include "renderer_util.h"
#include "sounder.h"
#include "controller.h"

#include "renderer_sdl.h"
#include "renderer_x11.h"
#include "renderer_ggi.h"
#include "renderer_w.h"

/* exports */
/* imports */
extern void	fbinit(void);

/* globals */
struct Renderer renderers[] = {
#ifdef HAVE_X
     { "x11", "X11 renderer", "linux",
       0,
       InitDisplayX11, UpdateDisplayX11, UpdateColorsX11, 0, 
       SaveScreenshotX11 },
     { "diff", "differential X11 renderer", "linux",
       RENDERER_DIFF,
       InitDisplayX11, UpdateDisplayX11, UpdateColorsX11, 0, 
       SaveScreenshotX11 },
     { "old", "old X11 renderer (tile-based)", "linux",
       RENDERER_OLD,
       InitDisplayX11, UpdateDisplayOldX11, UpdateColorsX11, 0,
       SaveScreenshotX11 },
#endif /* HAVE_X */
#ifdef HAVE_LIBSDL
     { "sdl", "SDL renderer", "sdl",
       RENDERER_NON_X11,
       InitDisplaySDL, UpdateDisplaySDL, 
       UpdateColorsSDL, ToggleFullScreenSDL, SaveScreenshotSDL },
#endif /* HAVE_LIBSDL */
#ifdef HAVE_GGI
     { "ggi", "GGI renderer", "linux",
       RENDERER_NON_X11,
       InitDisplayGGI, UpdateDisplayGGI, UpdateColorsGGI, 0, 0 },
#endif /* HAVE_GGI */
#ifdef HAVE_W
     { "w", "W renderer", "linux",
       RENDERER_NON_X11,
       InitDisplayW, UpdateDisplayW, UpdateColorsW, 0, 0 },
#endif /* HAVE_W */
     { "auto", "Choose one automatically", "linux",
       0,
       InitDisplayAuto, 0, 0, 0, 0 },
     { "none", "Don't draw anything", "linux",
       0,
       InitDisplayNone, UpdateDisplayNone, UpdateColorsNone, 0, 0 },
     { 0, 0, 0, 0, 0, 0, 0, 0, 0 }
}, *renderer = 0;

struct RendererConfig renderer_config = {
     /* int	inroot */
     0,
     /* int	sticky_keys */
     0,
     /* char	*geometry */
     0,
     /* char    *display_id */
     0
};

/* global renderer data */
struct RendererData renderer_data;

int
InitDisplayAuto(int argc, char **argv)
{
     char *rendname = 0;

#ifdef HAVE_W
     {
	  struct stat statbuf;
	  if (getenv ("WDISPLAY") ||
	      (! access ("/tmp/wserver", F_OK | W_OK) &&
	       (! stat ("/tmp/wserver", &statbuf)) &&
	       S_ISSOCK (statbuf.st_mode)))
	  {
	       rendname = "w";
	  }
     }
#endif
     if (getenv ("DISPLAY"))
     {
#ifdef HAVE_X
	  rendname = "x11";
#endif
#ifdef HAVE_GGI
	  if (! rendname)
	       rendname = "ggi";
#endif
#ifdef HAVE_W
	  if (! rendname)
	       rendname = "w";
#endif
     }
     else if (! rendname)
     {
#ifdef HAVE_GGI
	  rendname = "ggi";
#endif
#ifdef HAVE_W
	  if (! rendname)
	       rendname = "w";
#endif
     }
     if (! rendname)
     {
	  rendname = "none";
	  fprintf (stderr,
		   "======================================================\n"
		   "Warning: No Suitable Renderer Detected\n"
		   "\n"
		   "The `%s' fall-back renderer will be used.\n"
		   "\n"
		   "To avoid this in the future, install additional\n"
		   "libraries and recompile, or use the --renderer=...\n"
		   "option to specify a renderer explicitly.\n"
		   "\n"
		   "The option --help=options lists available renderers.\n"
		   "======================================================\n",
		   rendname);
     }
     for (renderer = renderers; renderer->name; renderer++)
	  if (! strcmp (renderer->name, rendname))
	       break;
     if (! renderer)
     {
	  fprintf (stderr, "%s: unrecognized renderer name `%s'\n",
		   *argv, rendname);
	  return 1;
     }
     return renderer->InitDisplay (argc, argv);
}

int
InitDisplayNone(int argc, char **argv)
{
     fbinit ();
     return 0;
}

void
UpdateDisplayNone(void)
{
  unsigned static int frame;

  /* do audio update */
  sounder->UpdateAudio();
  frame++;
  frame = resynchronize(frame);
}

/* Update the colors on the screen if the palette changed */
void
UpdateColorsNone(void)
{
     /* no-op */
}
