/*
 * This file is part of the TuxNES project codebase.
 *
 * Please see the README and COPYING files for more information regarding
 * this project.
 *
 * $Id: sounder_oss.c,v 1.4 2004/08/23 23:54:58 cheako Exp $
 *
 * Description: Sends sounds to the oss device.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif
#ifdef HAVE_ENDIAN_H
# include <endian.h>
#elif defined HAVE_SYS_ENDIAN_H
# include <sys/endian.h>
#elif defined HAVE_MACHINE_ENDIAN_H
# include <machine/endian.h>
#elif defined HAVE_SYS_PARAM_H
# include <sys/param.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "globals.h"
#include "sound.h"
#include "sounder.h"
#include "sounder_oss.h"

#ifdef HAVE_LINUX_SOUNDCARD_H
#include <linux/soundcard.h>
#endif /* HAVE_LINUX_SOUNDCARD_H */

#ifdef HAVE_SYS_SOUNDCARD_H
#include <sys/soundcard.h>
#endif /* HAVE_SYS_SOUNDCARD_H */

/* sample format numbers */
#ifndef AFMT_MU_LAW
#  define AFMT_MU_LAW 1
#endif
#ifndef AFMT_U8
#  define AFMT_U8 8
#endif
#ifndef AFMT_S16_LE
#  define AFMT_S16_LE 16
#endif
#ifndef AFMT_S16_BE
#  define AFMT_S16_BE 32
#endif
#ifndef AFMT_S8
#  define AFMT_S8 64
#endif
#ifndef AFMT_U16_LE
#  define AFMT_U16_LE 128
#endif
#ifndef AFMT_U16_BE
#  define AFMT_U16_BE 256
#endif

/* table of sample formats, terminated by { 0, ... } */
struct SampleFormat SampleFormatsOSS[] = {
     { "mu8",    "8-bit Mu-Law encoded *tested, imperfect",
       AFMT_MU_LAW, 1, 0, 0 },
     { "8",      "8-bit unsigned",
       AFMT_U8, 1, 0, 0 },
     { "8s",     "8-bit signed",
       AFMT_S8, 1, 1, 0 },
     { "16",     "16-bit signed",
#if BYTE_ORDER == BIG_ENDIAN
       AFMT_S16_BE, 2, 1, 0
#else
       AFMT_S16_LE, 2, 1, 0
#endif
     },
     { "16u",    "16-bit unsigned",
#if BYTE_ORDER == BIG_ENDIAN
       AFMT_U16_BE, 2, 0, 0
#else
       AFMT_U16_LE, 2, 0, 0
#endif
     },
     { "le16",   "16-bit signed (little-endian)",
       AFMT_S16_LE, 2, 1, 1 },
     { "le16u",  "16-bit unsigned (little-endian)",
       AFMT_U16_LE, 2, 0, 1 },
     { "be16",   "16-bit signed (big-endian)",
       AFMT_S16_BE, 2, 1, 2 },
     { "be16u",  "16-bit unsigned (big-endian)",
       AFMT_U16_BE, 2, 0, 2 },
     { 0, 0, 0, 0, 0, 0 }
};

int
InitAudioOSS(int argc, char **argv)
{
  /* Open an audio stream */
      if ((audiofd = open (sound_config.audiofile,
                           O_CREAT | O_WRONLY | O_APPEND, 0666 & ~mask)) < 0)
        {
          perror (sound_config.audiofile);
          return 1;
        }
      else
        {
#ifdef SNDCTL_DSP_RESET
          int desired_fragmentsize = 0; 
          if (! ioctl (audiofd, SNDCTL_DSP_RESET)) {
            int desired_audiorate;
            int desired_audiostereo;

            desired_audiorate = sound_config.audiorate;
            if (ioctl (audiofd, SNDCTL_DSP_SPEED, &sound_config.audiorate))
              perror (sound_config.audiofile);


            if (sound_config.audiorate != desired_audiorate)
              {
                fprintf (stderr, "%s: wanted %d Hz, got %d Hz\n",
                         sound_config.audiofile,
                         desired_audiorate,
                         sound_config.audiorate);
              }
            desired_audiostereo = sound_config.audiostereo;
            if ( desired_audiostereo == 1 ) {
                fprintf( stderr, "Stereo desired, but unsupported.  Mono forced. - pez\n");
                desired_audiostereo = 0;
            }
            if (ioctl (audiofd, SNDCTL_DSP_STEREO, &sound_config.audiostereo))
              perror (sound_config.audiofile);
            if (sound_config.audiostereo != desired_audiostereo)
              fprintf (stderr, "%s: wanted %s, got %s (expect failure)\n", sound_config.audiofile,
                       desired_audiostereo ? "stereo" : "monaural",
                       sound_config.audiostereo ? "stereo" : "monaural");
            if (ioctl (audiofd, SNDCTL_DSP_SAMPLESIZE, &sample_format->number))
              perror (sound_config.audiofile);
          }
#endif /* SNDCTL_DSP_RESET */
#ifdef SNDCTL_DSP_SETFRAGMENT
            /* this is necessary to get the sound out fast enough so that
             * the sound is in sync with the game
             * for performance issues, I believe a < 1/30 second delay is
             * very good, this will be tweaked for optimum performance
             */
            /* using samples_per_vsync as a temp */
            samples_per_vsync = sound_config.audiorate *
	     sample_format->bytes_per / 30;
            while ( samples_per_vsync ) {
                   ++desired_fragmentsize;
                   samples_per_vsync >>= 1;
            }
            if ( ioctl(audiofd, SNDCTL_DSP_SETFRAGMENT, &desired_fragmentsize) )
                perror (sound_config.audiofile);
            if ( verbose ) {
                  ioctl( audiofd, SNDCTL_DSP_GETBLKSIZE, &buf_size );
                fprintf(stderr, "BufSize: %u bytes\n",
                        (1 << desired_fragmentsize) );
            }
#endif /* SNDCTL_DSP_SETFRAGMENT */
          if (verbose)
            fprintf (stderr,
                     "Appending %d Hz %s samples to %s\n",
                     sound_config.audiorate,
                     sample_format->fullname
                     ? sample_format->fullname
                     : "(unknown sample format)",
                     sound_config.audiofile);
        }
   return InitAudioCore (argc, argv);
}


void
UpdateAudioOSS(void) /* called freq times a sec */
{
    static  unsigned char *buf;
#ifdef SNDCTL_DSP_GETODELAY
    static   unsigned char skip_count = 0;
#endif
    ProgressiveSoundUpdate( 1 );
    if ( audiofd < 0 ) return;
#ifdef SNDCTL_DSP_GETODELAY
    /* This is used to check sync every 0x0F frames instead of every frame.
     * Checking it every frame really slows the emulator down.
     * */
    if ( skip_count ) {
        --skip_count;
        return;
    }
    if (sound_config.max_sound_delay > 0.0) {
        static unsigned long frame_count = 0;
        static int maxdelay = 0;
               int odelay;
        ++frame_count;
        if ( (frame_count & 0x0F) == 0x0F && /* check it every few frames */
             !ioctl (audiofd, SNDCTL_DSP_GETODELAY, &odelay)) {
            if (odelay > maxdelay) {
                maxdelay = odelay;
            }
            if (odelay * 1.0 / (sound_config.audiorate *
				sample_format->bytes_per) >=
                sound_config.max_sound_delay) {
                skip_count = odelay * UPDATE_FREQ / (sound_config.audiorate 
                             * sample_format->bytes_per) + 1;
                  if (verbose) {
                       fprintf (stderr, 
                             "Warning: %.3f sec sound delay, resynchronizing\n",
                             odelay * 1.0 / (sound_config.audiorate 
                             * sample_format->bytes_per));
                       fprintf (stderr, "Skipping %u frames.\n", skip_count );
                  }
                return;
            }
        }
    }
#endif /* SNDCTL_DSP_GETODELAY */
    if ( halfspeed ) {
        for ( buf = audio_buffer;
              buf < (audio_buffer + samples_per_vsync);
              buf += sample_format->bytes_per ) {
            write( audiofd, buf, sample_format->bytes_per);
            write( audiofd, buf, sample_format->bytes_per);
        }
    } else if ( doublespeed ) {
        for ( buf = audio_buffer;
              buf < (audio_buffer + samples_per_vsync);
              buf += (doublespeed * sample_format->bytes_per) ) {
            write (audiofd, buf, sample_format->bytes_per);
        }
    } else
        write ( audiofd, audio_buffer, samples_per_vsync );
}

void ShutdownAudioOSS(void)
{
    close(audiofd);
}
